#include <iomanip>
#include <vector>
#include <algorithm>
#include <iostream>
#include <fstream>
//#include <string>
#include <sstream>
#include <stdio.h>
#include <cmath>
#include <ctime>
#include <cstdlib>
#include <sys/stat.h> 
#include <sys/time.h> 
#include <gsl/gsl_sf.h> 
#include <gsl/gsl_interp.h>
#include <gsl/gsl_spline.h>
#include <gsl/gsl_integration.h>
#include <getopt.h> 


#include "math/complex.hh"
#include "math/mellin_Dk.hh"
#include "math/integration.hh"
#include "math/special_functions.hh"
#include "math/chebyshev.hh"
#include "CFs_pointlike/CFs_pointlike.hh"
#include "NLO_finite-mt/bdv.hh"
#include "NNLO_finite-mt/nnlo_finite-mt.hh"
#include "NNLO_finite-mt/nlo_finite-mt.hh"
#include "ggHiggs.hh"
#include "msbar_mass.hh"
#include "parameters.hh"
#include "anomalous_dimensions/gammaLO.hh"
#include "anomalous_dimensions/gammaNLO.hh"
#include "anomalous_dimensions/gammaNNLO.hh"

#include <cuba.h>


using namespace std;




namespace ggHiggs {

#include "deltaEW.cc"

const int _nf = 5;


  // PRIVATE flags and variables ( don't change here !!! )
  double mtop = 172.5;  // Pole mass
  double mtmt = 162.7;  // MSbar mass
  double mbot = -1; // negative mass means no bottom running in the loop
  double mcha = -1; // negative mass means no charm running in the loop
vector<double> qmasses;
dcomplex mH2;
int order = 1;
//int PDForder = 1;
bool finite_mt = false;
SmallxCoeffs *sxc;
double _muF_mu_ratio  = 1.;
double _muR_muF_ratio = 1.;
  double muF, muR, muM, muFpdf, mut;
bool __exact__ = false;
bool __soft_part__ = true;
bool __hard_part__ = true;
int __log_type__ = 0; // 0:D, 1:Dhat, 2:Dlog
bool __Nsoftcoll__ = false;
bool __NNLL__ = false;
int APtype = 0;
bool APimprExponent = true;
bool MSbarMass = false;
bool MSbar = true;
bool quiet = true;
double INTEGRAL_PRECISION = 15.e-3;
bool exactmuRdep = true;
bool exactmuFdep = true;
bool computesoft = true;
double _integration_v_;

  bool _gg_channel_ = true;
  bool _qg_channel_ = true;
  bool _qq_channel_ = true;

  bool newevolmuFmuR = 1;







// alpha_s running from ihixs
double alphas_ihixs(double mu, int porder, double as0, double mu0=mZ) {
  // Nf = 5 **************** <---------
  double bb0 = 1.916666667;
  double bb1 = 1.260869565;
  double bb2 = 1.474788647;
  double bb3 = 9.835916431;
  //
  double as;
  as0 /= M_PI;
  //double as0 = asZ/M_PI;
  //double as0=asZ/3.1415926;
  //double Mz=91.187;
  //
  int N = 100000;
  //double step = 2.*log(mu/mZ)/N, incr;
  //double step = 2.*log(mu/Mz)/N, incr;
  double step = 2.*log(mu/mu0)/N, incr;
  for (int i=0; i<N; i++) {
    if (porder >= 3)
      incr = step*bb0*as0*as0*(1+bb1*as0 + bb2*as0*as0 + bb3*as0*as0*as0 );
    else if (porder == 2)
      incr = step*bb0*as0*as0*(1+bb1*as0 + bb2*as0*as0 );
    else if (porder == 1)
      incr = step*bb0*as0*as0*(1+bb1*as0);
    else if (porder == 0)
      incr = step*bb0*as0*as0;
    as = as0 - incr;
    as0 = as;
  }
  return as*M_PI;
}







  double b0 = beta0(_nf);
  double b1 = beta1(_nf);
  double b2 = beta2(_nf);


  double e01=CA, e00, e11, e22;
  void initGammaCoeffs(bool gamma_NLL, int nf) {
    if(gamma_NLL) {
      e00 = (-11.*CA+2.*nf*(2.*CF/CA-1.))/12.;
      e11 = nf*(26.*CF-23.*CA)/36.;
      e22 = CA*( (18.*ZETA2-71.)*(CA-2*CF)*nf + CA*CA*(54*ZETA3+99.*ZETA2-395) ) /108.;
    }
    else e00=e11=e22=0;
  }
dcomplex gamma0poles(dcomplex N, int nf) {
  return e01/(N-1) + e00;
}
dcomplex gamma1poles(dcomplex N, int nf) {
  return e11/(N-1);
}
dcomplex gamma2poles(dcomplex N, int nf) {
  return e22/(N-1)/(N-1);
}
dcomplex FO_1_sx_NLL_RC(dcomplex N, double mu) {
  double lf = -2.*log(_muF_mu_ratio);
  double c10 = sxc->c10(mu,mtop);
  double c1 = 2*c10;
  c1 += 2.*lf;
  return c1*gamma0poles(N,_nf);
}
dcomplex FO_2_sx_NLL_RC(dcomplex N, double mu) {
  double lf = -2.*log(_muF_mu_ratio);
  double c10 = sxc->c10(mu,mtop);
  double c20 = sxc->c20(mu,mtop);
  double c11 = sxc->c11(mu,mtop);
  double c1 = 2*c10;
  double c2 = 2.*c20+c11;
  c20 += (c1+lf)*lf/2.;
  c2  += (c1+lf)*lf*2.;
  c1  += 2*lf;
  dcomplex g0 = gamma0poles(N,_nf);
  dcomplex g1 = gamma1poles(N,_nf);
  return c2*g0*g0 - 2*c20*b0*g0 + c1*g1 ;
}
dcomplex FO_3_sx_NLL_RC(dcomplex N, double mu) {
  double lf = -2.*log(muF/muM);
  double c10 = sxc->c10(mu,mtop);
  double c20 = sxc->c20(mu,mtop);
  double c11 = sxc->c11(mu,mtop);
  double c30 = sxc->c30(mu,mtop);
  double c21 = sxc->c21(mu,mtop);
  if(MSbar) c30 += 8./3.*ZETA3;
  double c1 = 2*c10;
  double c2 = 2.*c20+c11;
  double c3 = 2.*(c30+c21);
  // scale dependence
  c30 += c20*lf + c1*lf*lf/4. + lf*lf*lf/6.;
  c21 += (c2-c20)*lf + 3.*c1*lf*lf/4. + lf*lf*lf/2.;
  c20 += (c1+lf)*lf/2.;
  c2  += (c1+lf)*lf*2.;
  c1  += 2*lf;
  dcomplex g0 = gamma0poles(N,_nf);
  dcomplex g1 = gamma1poles(N,_nf);
  dcomplex g2 = gamma2poles(N,_nf);
  dcomplex hard = (c3*g0*g0*g0 - (3*c30+c21)*2*b0*g0*g0 + 4*c30*b0*b0*g0
		  + 2*c2*g0*g1 - 4*c20*b0*g1 + c1*g2 );
  //double lrf = 2.*log(_muR_muF_ratio);
  //hard += 4*b0*FO_2_sx_NLL_RC(N,mu)*lrf;
  return hard;
}
double MomConsCoeff[3] = {0,0,0};
void ComputeMomConsCoeff(double mu) {
  MomConsCoeff[0] = real(FO_1_sx_NLL_RC(2,mu) - 2.*FO_1_sx_NLL_RC(3,mu) + FO_1_sx_NLL_RC(4,mu));
  MomConsCoeff[1] = real(FO_2_sx_NLL_RC(2,mu) - 2.*FO_2_sx_NLL_RC(3,mu) + FO_2_sx_NLL_RC(4,mu));
  MomConsCoeff[2] = real(FO_3_sx_NLL_RC(2,mu) - 2.*FO_3_sx_NLL_RC(3,mu) + FO_3_sx_NLL_RC(4,mu));
}
void ResetMomConsCoeff() {
  MomConsCoeff[0] = 0;  MomConsCoeff[1] = 0;  MomConsCoeff[2] = 0;
}




double Dk(int k, double z) {
  return pow(log(1-z),k)/(1.-z);
}
double Dtk(int k, double z) {
  return pow(log(1-z)-0.5*log(z),k)/(1.-z);
}
double DMPk(int k, double z) {
  return pow(log(log(1./z)),k)/log(1./z);
}
double mell_const(int k, double fake=0) {
  return DG(k+1,1.)/(k+1.);
}




// perturbative expansions of constants and logs
double _cd1, _cd2, _cd3;
void init_delta_const() {
  _cd1 = 2.*CA * ZETA2;
  _cd2 = ( 837./16. + 67./2.*ZETA2 - 9./20.*ZETA2*ZETA2 - 165./4.*ZETA3
	   + _nf  * ( 5./6.*ZETA3 - 5./3.*ZETA2 - 247./36. ) );
  _cd3 = 1740.02 - 131.928*_nf + 1.75719*_nf*_nf; // new coefficient
}
double _W1, _W2, _W3;
double _W1t, _W2t, _W3t;
void init_Wilson(double mu) {
  double lHt = 2.*log(mu/mtop);
  _W1 = 11./2.;
  _W2 = 121./16. + 2777./144. + 19./8.*lHt + (-67./48. + 2./3.*lHt)*_nf;
  // this was MSbar
  //_W3 = 40.6488 + 40.1458*lHt + 10.941*lHt*lHt; // write full expression .... done.
  //_W3 = (-1792967./20736. +  897943./4608.*ZETA3 + 5347./288.*lHt + 209./32.*lHt*lHt
  //     + _nf * (493./10368. + 209./54.*lHt + 23./16.*lHt*lHt - 110779./6912.*ZETA3)
  //	 + _nf*_nf * (-6865./15552. + 77./864.*lHt - lHt*lHt/9. ) );
  // this is OS
  _W3 = (-1661639./20736. +  897943./4608.*ZETA3 + 6715./288.*lHt + 209./32.*lHt*lHt
	 + _nf * (18925./10368. + 281./54.*lHt + 23./16.*lHt*lHt - 110779./6912.*ZETA3)
	 + _nf*_nf * (-6865./15552. + 77./864.*lHt - lHt*lHt/9. ) );
  if(MSbarMass) _W3 += (19./8. + 2./3.*_nf)*(-8./3.-4*log(mut/mtmt));  // conversion to MSbar mass
  //cout << _W1/pow(M_PI,1) << endl;
  //cout << _W2/pow(M_PI,2) << endl;
  //cout << _W3/pow(M_PI,3) << endl;
}
void init_Wilson_mt(double mu) {
  _W1t = G2l(mu,mtop,mbot,mcha);
  _W2t = gg_nnlo_mt_delta(mu, mtop, 1) - _cd2 - _cd1*_W1t;
  _W3t = _W3; // pointlike value
  if(mu > 2*mtop) _W2t = _W2; // pointlike value
}
//
#include "ggHiggs_Mellin.cc"
//
void init_consts(double mH) {
  if(MSbarMass) mtop = runmass(mtmt, alphas(mtmt)/M_PI, alphas(mut)/M_PI, 3, _nf);
  //initBDV(mH, mtop, mbot, mcha);
  init_delta_const();
  init_Wilson(mH);
  init_Wilson_mt(mH);
}
void SetScales(double mH) {
  muM = mH;
  muFpdf = _muF_mu_ratio * mH;
  if(exactmuFdep) muF = mH;
  else muF = _muF_mu_ratio * mH;
  if(exactmuRdep) muR = muF;
  else muR = _muR_muF_ratio * muF;
  mut = _muR_muF_ratio * _muF_mu_ratio * mH;
  initResConsts(mH);
}
void SetScaleDependence(bool exact_muR, bool exact_muF) {
  exactmuRdep = exact_muR;
  exactmuFdep = exact_muF;
}




//
double APimprovement(double z) {
  double res = 0;
  if(APtype== 0) res = 1;                                       // Ag(1)
  if(APtype== 1) res = z;                                       // Ag expanded to first order
  if(APtype== 2) res = (2-3*z+2*z*z);                           // full Ag expanded to second order
  if(APtype==-1) res = (1./z-2+3*z-2*z*z+z*z*z);                // full Ag
  if(APtype==-2) res = (3*z-2*z*z+z*z*z/2);                     // soft Ags
  if(APtype==-3) res = 1./z;                                    // small-x
  return res;
}







double FO_1_log_PL(double z, double D(int,double)) {
  return C1log_pointlike_gg(z, D, _nf);
}
double FO_2_log_PL(double z, double D(int,double)) {
  return C2log_pointlike_gg(z, D, _nf);
}
double FO_3_log_PL(double z, double D(int,double)) {
  return C3log_pointlike_gg(z, D, _nf);
}









// Fixed-order (NLO)
///////////////////////////////////////
double FO_1_log(double z, double D(int,double)) {
  double lf = -2.*log(_muF_mu_ratio);
  return FO_1_log_PL(z,D) + 2*CA*D(0,z)*lf;
}
double FO_1_log_APimproved(double z, double D(int,double)) {
  return APimprovement(z)*FO_1_log(z,D);
}
double FO_1_Nsoftcoll_impr(double z) {
  return 2.*CA*(gsl_sf_psi(1.)-log(log(1./z)));
}
double FO_1_delta_const() {
  double W1 = (finite_mt ? _W1t : _W1);
  double c = _cd1 + W1;
  //double lrf = 2.*log(_muR_muF_ratio);
  //c += 2*b0*lrf;
  return c;
}
// finite mt R_gg term
double Rgg_zspace(double z, double mu) {
  double z0 = 2e-4; //  might not be super precise, Rgg is very unstable at small x
  double c1 = sxc->c1marzani(mu,mtop);
  //double rgg = (z<z0 ? (c1+2.*log(z))/z : Rgg(z,mu*mu/z)); // old BDV fortran implementation
  double rgg = (z<z0 ? (c1+2.*log(z))/z : Rgg(z,mH2,qmasses,_integration_v_));
  return rgg;
}
double Rqg_zspace(double z, double mu) {
  double z0 = 1e-6;  // it behaves better than Rgg (no H3 function in it)
  double c1 = sxc->c1marzani(mu,mtop);
  //double rqg = (z<z0 ? (c1/2. + log(z))/z : Rqg(z,mu*mu/z)); // old BDV fortran implementation
  double rqg = (z<z0 ? (c1/2. + log(z))/z : Rqg(z,mH2,qmasses,_integration_v_));
  return CF*rqg;
}
double Rqqbar_zspace(double z, double mu) {
  // This has no numerical instabilities (no numerical integration)
  //return Rqqbar(z,mu*mu/z); // old BDV fortran implementation
  return Rqqbar(z,mH2,qmasses);
}
double FO_1_nonlog(double z, double mu) {
  double lf = -2.*log(_muF_mu_ratio);
  double smallx = (finite_mt ? CA*Rgg_zspace(z, mu) : - 11./2.*pow(1.-z,3.)/z );
  double nonlog = CA * 2.*( (1./z-2.+z*(1.-z))*(2.*log(1.-z)-log(z)) -log(z)/(1.-z) ) + smallx;
  return nonlog + 2.*CA*(1./z-2.+z-z*z)*lf;
}
double FO_1_hard(double z, double mu) {
  double lf = -2.*log(_muF_mu_ratio);
  double c1 = sxc->c1marzani(mu,mtop);
  c1 += 2.*lf;
  double hard = c1*e01/z;// + c1*e00;  // delta term in z-space, cancelled by damping
  return (hard - 12.*MomConsCoeff[0]) * pow(1-z, 2);
}
double FO_1_qg(double z, double mu) {
  double lf = -2.*log(_muF_mu_ratio);
  double Pgq = CF*(1.+pow(1.-z,2))/z/2.;
  double smallx = (finite_mt ? Rqg_zspace(z, mu) : 2./3.*z - pow(1.-z,2)/z );
  return Pgq*(2.*log(1-z)-log(z)+lf) + smallx;
}
double FO_1_qqb(double z, double mu) {
  return (finite_mt ? Rqqbar_zspace(z, mu) : 32./27.*pow(1-z,3)/z );
}
//




// Fixed-order (NNLO)
///////////////////////////////////////
dcomplex FO_2_log(dcomplex N) {
  double W1 = (finite_mt ? _W1t : _W1);
  return FO_2_log_PL(N) + W1*S1(N);
}
double FO_2_log(double z, double D(int,double)) {
  double lf = -2.*log(_muF_mu_ratio);
  double W1 = (finite_mt ? _W1t : _W1);
  double Clog = FO_2_log_PL(z,D) + W1*FO_1_log_PL(z,D);
  Clog += ( 108*lf * D(2,z)
	    + (36.*lf*lf - (33.-2.*_nf)*lf) * D(1,z)
	    + (-(33.-2.*_nf)/4.*lf*lf + (2.*CA*W1 + 67./2.-5./3.*_nf-5.*CA*CA*ZETA2)*lf) * D(0,z) );
  //double lrf = 2.*log(_muR_muF_ratio);
  //Clog += 3*b0*FO_1_log(z,D)*lrf;
  return Clog;
  //if(__exact__ && finite_mt) return gg_nnlo_mt_distr(z, mu, mtop, _muF_mu_ratio); // no longer needed
}
dcomplex FO_2_log_APimprovement(dcomplex N, void *par) {
  return 0.5*(pow(APimprovement(S1,N),2) - APimprovement(S1_sq,N));
}
double FO_2_log_APimprovement_soft1(double z) {
  double piece = (Li3(z) - ZETA3  + 0.5*log(z)*(-ZETA2 + 1./6.*pow(log(z),2) - Li2(z) - 0.5*log(z)*log(1-z)))/(1.-z);
  return 8*CA*CA*z*piece;
}
double FO_2_log_APimprovement_soft2(double z) {
  double piece = (Li3(z) - ZETA3  + 0.5*log(z)*(-ZETA2 + 1./6.*pow(log(z),2) - Li2(z) - 0.5*log(z)*log(1-z)))/(1.-z);
  return 8*CA*CA*(-4.+4.*z+4.*ZETA2-12.*z*ZETA2
		  +8.*log(1.-z) + 2.*(-1.+ZETA2+z*(3.+ZETA2))*log(z) +2.*z*pow(log(z),2)
		  -1./12.*(1.+z)*pow(log(z),3) - 4.*log(1-z)* (ZETA2+z*(2.+ZETA2)+2.*(-1+z)*log(z))
		  +2.*pow(log(1-z),2)*(-4.+4.*z+log(z)+z*log(z)) + 2.*(1.+z)*(2.+2.*log(1-z)-log(z))*Li2(z)
		  +4.*(1+z)*Li3(1-z) + 2.*(1+z)*Li3(z) - 2.*(1+z)*ZETA3
		  + (2.-3.*z+2.*z*z)*piece);
}
double FO_2_log_APimproved(double z, double D(int,double)) {
  double res = APimprovement(z)*FO_2_log(z,D);
  //if(APimprExponent) res += InverseMellin(z,FO_2_log_APimprovement,NULL);
  if(APimprExponent) {
    if     (APtype == 1 && __log_type__ == 1) res += FO_2_log_APimprovement_soft1(z);
    else if(APtype == 2 && __log_type__ == 1) res += FO_2_log_APimprovement_soft2(z);
    else res += InverseMellin(z,FO_2_log_APimprovement,NULL);
  }
  return res;
}
dcomplex FO_2_Nsoftcoll_impr(dcomplex N, void *par) {
  return 2.*CA*log(N)/N*S1(N);
}
double FO_2_Nsoftcoll_impr(double z) {
  return InverseMellin(z,FO_2_Nsoftcoll_impr,NULL);
}
double FO_2_delta_const(double mu) {
  double lf = -2.*log(_muF_mu_ratio);
  double W1 = (finite_mt ? _W1t : _W1);
  double W2 = (finite_mt ? _W2t : _W2);
  double c = _cd2 + W2 + _cd1*W1;
  c += -18.*ZETA2*lf*lf + (-3./4. + 171./2.*ZETA3 + 1./4.*_nf - (33.-2.*_nf)/2.*ZETA2 - W1*(33.-2.*_nf)/12.)*lf;
  //if(finite_mt) c = gg_nnlo_mt_delta(mu, mtop, _muF_mu_ratio); // no longer needed (checked)
  //
  //double lrf = 2.*log(_muR_muF_ratio);
  //double muRF = _muR_muF_ratio;  _muR_muF_ratio = 1.;
  //c += (2*b1+3*b0*FO_1_delta_const())*lrf + 3*b0*b0*lrf*lrf;
  //_muR_muF_ratio = muRF;
  return c;
}
// sigma(z) = z C(z)
// C(z) = [D(z)]_+ + nonlog(z)
// sigma(z) = [D(z)]_+ - (1-z)D(z) + z*nonlog(z)
//          = [D(z)]_+ + reg(z)
// nonlog(z) = [ reg(z) + (1-z)D(z) ] /z
double FO_2_nonlog(double z, double mu) {
  double nonlog;
  if(!finite_mt) nonlog = C2reg_pointlike_gg(z, _nf, _muF_mu_ratio) + _W1 * C1reg_pointlike_gg(z, _nf, _muF_mu_ratio);
  else {
    double lf = -2.*log(_muF_mu_ratio);
    double c1 = sxc->c1marzani(mu,mtop);
    double c2 = sxc->c2marzani(mu,mtop);
    c2  += (c1+lf)*lf*2.;
    if(mu<2*mtop)
      nonlog = ( gg_nnlo_mt_reg(z, mu, mtop, _muF_mu_ratio) + smallx_match_z(z, c2) + (1.-z)*gg_nnlo_mt_distr(z, mu, mtop, _muF_mu_ratio) ) /z;
    else {
      //nonlog = ( gg_nnlo_infmt_reg(z, mu, mtop, _muF_mu_ratio) + smallx_match_z(z, c2) + (1.-z)*gg_nnlo_infmt_distr(z, mu, mtop, _muF_mu_ratio) ) /z;
      nonlog = ( gg_nnlo_infmt_reg(z, mu, mtop, _muF_mu_ratio) + (1.-z)*gg_nnlo_infmt_distr(z, mu, mtop, _muF_mu_ratio) ) /z;
      //finite_mt = false;  nonlog = ( C2reg_pointlike(z, _nf, _muF_mu_ratio) + (1.-z)*FO_2_log(z,Dk) ) /z;  finite_mt = true;
      double nonlog1 = CA * 2.*( (1./z-2.+z*(1.-z))*(2.*log(1.-z)-log(z)) -log(z)/(1.-z) );
      nonlog += (_W1t-_W1)*nonlog1;
    }
  }
  //double lrf = 2.*log(_muR_muF_ratio);
  //nonlog += 3*b0*FO_1_nonlog(z,mu)*lrf;
  return nonlog;
}
//
double FO_2_hard(double z, double mu) {
  //double hard = SF(z, mu, FO_2_sx_NLL_RC); // ok checked
  //double lrf = 2.*log(_muR_muF_ratio);
  //hard += 3*b0*FO_1_hard(z,mu)*lrf;
  // z-space version: faster
  double lf = -2.*log(_muF_mu_ratio);
  double c10 = sxc->c10(mu,mtop);
  double c20 = sxc->c20(mu,mtop);
  double c11 = sxc->c11(mu,mtop);
  double c1 = 2*c10;
  double c2 = 2.*c20+c11;
  c20 += (c1+lf)*lf/2.;
  c2  += (c1+lf)*lf*2.;
  c1  += 2*lf;
  double hard = -c2*e01*e01*log(z)/z + (2*c2*e00*e01-2*c20*b0*e01+c1*e11)/z;
  //
  return (hard - 12.*MomConsCoeff[1]) * pow(1-z, 2);
}
double FO_2_qg(double z, double mu) {
  double res;
  if(!finite_mt) res = C2reg_pointlike_qg(z, _nf, _muF_mu_ratio) + _W1*C1reg_pointlike_qg(z, _nf, _muF_mu_ratio);
  else {
    double lf = -2.*log(_muF_mu_ratio);
    double c10 = sxc->c10(mu,mtop);
    double c20 = sxc->c20(mu,mtop);
    double c11 = sxc->c11(mu,mtop);
    double c1 = 2*c10;
    double c2 = 2.*c20+c11;
    c2  += (c1+lf)*lf*2.;
    c20 += (c1+lf)*lf/2.;
    if(mu<2*mtop)
      res = ( qg_nnlo_mt(z, mu, mtop, _muF_mu_ratio) + smallx_match_z(z, (c2-c20)*CF/CA) ) /z;
    else {
      res = qg_nnlo_infmt(z, mu, mtop, _muF_mu_ratio) /z;
      double Pgq = CF*(1.+pow(1.-z,2))/z/2.;
      double res1 = Pgq*(2.*log(1-z)-log(z));
      res += (_W1t-_W1)*res1;
    }
  }
  return res;
}
double FO_2_qqb(double z, double mu) {
  double res;
  if(!finite_mt) res = C2reg_pointlike_qqb(z, _nf, _muF_mu_ratio) + _W1*C1reg_pointlike_qqb(z, _nf, _muF_mu_ratio) ;
  else {
    double lf = -2.*log(_muF_mu_ratio);
    double c10 = sxc->c10(mu,mtop);
    double c20 = sxc->c20(mu,mtop);
    double c11 = sxc->c11(mu,mtop);
    double c1 = 2*c10;
    double c2 = 2.*c20+c11;
    c2  += (c1+lf)*lf*2.;
    c20 += (c1+lf)*lf/2.;
    if(mu<2*mtop)
      res = ( qqb_nnlo_mt(z, mu, mtop, _muF_mu_ratio) + smallx_match_z(z, (c2-2*c20)*pow(CF/CA,2)) ) /z;
    else {
      res = qqb_nnlo_infmt(z, mu, mtop, _muF_mu_ratio) /z;
    }
  }
  return res;
}
double FO_2_qq(double z, double mu) {
  double res;
  if(!finite_mt) res = C2reg_pointlike_qq(z, _nf, _muF_mu_ratio);
  else {
    double lf = -2.*log(_muF_mu_ratio);
    double c10 = sxc->c10(mu,mtop);
    double c20 = sxc->c20(mu,mtop);
    double c11 = sxc->c11(mu,mtop);
    double c1 = 2*c10;
    double c2 = 2.*c20+c11;
    c2  += (c1+lf)*lf*2.;
    c20 += (c1+lf)*lf/2.;
    if(mu<2*mtop)
      res = ( qq_nnlo_mt(z, mu, mtop, _muF_mu_ratio) + smallx_match_z(z, (c2-2*c20)*pow(CF/CA,2)) ) /z;
    else {
      res = qq_nnlo_infmt(z, mu, mtop, _muF_mu_ratio) /z;
    }
  }
  return res;
}
double FO_2_qqp(double z, double mu) {
  double res;
  if(!finite_mt) res = C2reg_pointlike_qqp(z, _nf, _muF_mu_ratio);
  else {
    double lf = -2.*log(_muF_mu_ratio);
    double c10 = sxc->c10(mu,mtop);
    double c20 = sxc->c20(mu,mtop);
    double c11 = sxc->c11(mu,mtop);
    double c1 = 2*c10;
    double c2 = 2.*c20+c11;
    c2  += (c1+lf)*lf*2.;
    c20 += (c1+lf)*lf/2.;
    if(mu<2*mtop)
      res = ( qqp_nnlo_mt(z, mu, mtop, _muF_mu_ratio) + smallx_match_z(z, (c2-2*c20)*pow(CF/CA,2)) ) /z;
    else {
      res = qqp_nnlo_infmt(z, mu, mtop, _muF_mu_ratio) /z;
    }
  }
  return res;
}





  

// Fixed-order (NNNLO)
///////////////////////////////////////
dcomplex FO_3_log(dcomplex N) {
  double W1 = (finite_mt ? _W1t : _W1);
  double W2 = (finite_mt ? _W2t : _W2);
  return FO_3_log_PL(N) + W1*FO_2_log_PL(N) + W2*S1(N);
}
double D0D1(double z, double D(int,double)) {
  return 3./2.*D(2,z) - ZETA2*D(0,z);
}
double D0D2(double z, double D(int,double)) {
  return 4./3.*D(3,z) - 2.*ZETA2*D(1,z) + 2.*ZETA3*D(0,z);
}
double D0D3(double z, double D(int,double)) {
  return 5./4.*D(4,z) - 3.*ZETA2*D(2,z) + 6.*ZETA3*D(1,z) - 6.*ZETA4*D(0,z);
}
double FO_3_log(double z, double D(int,double)) {
  double lf = -2.*log(muF/muM);
  double W1 = (finite_mt ? _W1t : _W1 );
  double W2 = (finite_mt ? _W2t : _W2 );
  double Clog = FO_3_log_PL(z,D) + W1*FO_2_log_PL(z,D) + W2*FO_1_log_PL(z,D);
  if(__NNLL__) {
    Clog -= 0.953938*D(0,z) *M_PI*M_PI*M_PI;
  }
  Clog += ( 0*lf * D(4,z)
	    + (16.*CA*CA*CA*lf*lf + 0*lf) * D(3,z)
	    + (4.*CA*CA*CA*lf*lf*lf - 18.*CA*CA*b0*lf*lf + 12.*CA*A1*lf) * D(2,z)
	    + (-4.*CA*CA*b0*lf*lf*lf + (8.*CA*A1+4.*CA*b0*b0+4.*CA*CA*W1-32.*CA*CA*CA*ZETA2)*lf*lf + (8*CA*B1-12*CA*b1)*lf) * D(1,z)
	    + ((2.*b0*b0-4.*CA*CA*CA*ZETA2)*lf*lf*lf + (CA*(4.*B1-5.*b1)-3.*CA*b0*W1-2.*A1*b0+6.*CA*CA*b0*ZETA2+32.*CA*CA*CA*ZETA3)*lf*lf
	       + (2.*A2+2.*W1*A1-4.*CA*A1*ZETA2)*lf) * D(0,z) );
  // adding contributions from interference with C2
  Clog += lf*(2.*CA*( 72. * D0D3(z,D)
		      + (2.*_nf - 33.) * D0D2(z,D)
		      + (67.+4*CA*W1 - 90.*ZETA2 - 10./3.*_nf) * D0D1(z,D)
		      + (-101./3. + 33.*ZETA2 + 351./2.*ZETA3 + _nf*(14./9.-2.*ZETA2)) * 2*D(1,z)
		      + (_cd2 + W2 + _cd1*W1 ) * D(0,z)
		      )
	      -2.*b0*( 72. * D(3,z)
		       + (2.*_nf - 33.) * D(2,z)
		       + (67.+4*CA*W1 - 90.*ZETA2 - 10./3.*_nf) * D(1,z)
		       + (-101./3. + 33.*ZETA2 + 351./2.*ZETA3 + _nf*(14./9.-2.*ZETA2)) * D(0,z)
		       ) );
  //double lrf = 2.*log(_muR_muF_ratio);
  //double muRF = _muR_muF_ratio;  _muR_muF_ratio = 1.;
  //Clog += (3*b1*FO_1_log(z,D)+4*b0*FO_2_log(z,D))*lrf + 6*b0*b0*FO_1_log(z,D)*lrf*lrf;
  //_muR_muF_ratio = muRF;
  return Clog;
}
dcomplex FO_3_log_APimprovement(dcomplex N, void *par) {
  double W1 = (finite_mt ? _W1t : _W1);
  double d1 = _cd1 + W1;
  return 1./6.*(pow(APimprovement(S1,N),3) - APimprovement(S1_cube,N))
    + APimprovement(S1,N)*APimprovement(S2,N) - APimprovement(S1_S2,N)
    + d1/2.*(pow(APimprovement(S1,N),2) - APimprovement(S1_sq,N))
    ;
}
double FO_3_log_APimproved(double z, double D(int,double)) {
  double res = APimprovement(z)*FO_3_log(z,D);
  if(APimprExponent) res += InverseMellin(z,FO_3_log_APimprovement,NULL);
  return res;
}
dcomplex FO_3_Nsoftcoll_impr(dcomplex N, void *par) {
  return 2.*CA*log(N)/N*(S2(N)+S1(N)*S1(N)/2.);
}
double FO_3_Nsoftcoll_impr(double z) {
  return InverseMellin(z,FO_3_Nsoftcoll_impr,NULL);
}
double FO_3_delta_const(double mu) {
  double lf = -2.*log(muF/muM);
  // not yet known in the full theory => use pointlike value.
  //double W1 = (finite_mt ? _W1t : _W1 );
  //double W2 = (finite_mt ? _W2t : _W2 );
  //double W3 = (finite_mt ? _W3t : _W3 );
  double W1 = _W1;
  double W2 = _W2;
  double W3 = _W3;
  //
  double c = W3 + W2*_cd1 + W1*_cd2 + _cd3;
  //c = (52.1348 - 1.97392*FO_1_delta_const()/M_PI) *M_PI*M_PI*M_PI;  // this corresponds to g0bar3 = 0 only for muF=mH
  if(__NNLL__) c = -FO_3_log(1, mell_const);
  //cout << c/M_PI/M_PI/M_PI << endl;
  //
  // factorization scale dependence
  c += (72.*ZETA3+18.*b0*ZETA2)*lf*lf*lf
    + (-4.*CA*A1*ZETA2-2.*b0*(B1-b1+6.*CA*CA*ZETA3)+b0*b0*(2.*CA*ZETA2+W1)-2.*CA*CA*W1*ZETA2+4.*CA*CA*CA*(ZETA2*ZETA2-6.*ZETA4))*lf*lf
    + (4.*CA*B1*ZETA2-6.*CA*b1*ZETA2-3.*W1*b1+2.*B2-2.*b2+2.*W1*B1+8.*CA*A1*ZETA3)*lf;
  // adding contributions from interference with C2
  c += lf*(2*CA*( 72. * 6*ZETA5
		  + (2.*_nf - 33.) * (-2*ZETA4)
		  + (67.+4*CA*W1 - 90.*ZETA2 - 10./3.*_nf) * ZETA3
		  + (-101./3. + 33.*ZETA2 + 351./2.*ZETA3 + _nf*(14./9.-2.*ZETA2)) * (-ZETA2)
		  )
	   -  2*b0* (_cd2 + W2 + _cd1*W1 ) );
  //double lrf = 2.*log(_muR_muF_ratio);
  //double muRF = _muR_muF_ratio;  _muR_muF_ratio = 1.;
  //c += (2*b2+3*b1*FO_1_delta_const()+4*b0*FO_2_delta_const(mu))*lrf + (7*b0*b1+6*b0*b0*FO_1_delta_const())*lrf*lrf + 4*b0*b0*b0*lrf*lrf*lrf;
  //_muR_muF_ratio = muRF;
  return c;
}
double FO_3_hard(double z, double mu) {
  //double hard = SF(z, mu, FO_3_sx_NLL_RC); // ok checked
  //double lrf = 2.*log(_muR_muF_ratio);
  //hard += FO_1_hard(z,mu) * (3*b1*lrf + 6*b0*b0*lrf*lrf);
  //
  double lf = -2.*log(muF/muM);
  double c10 = sxc->c10(mu,mtop);
  double c20 = sxc->c20(mu,mtop);
  double c11 = sxc->c11(mu,mtop);
  double c30 = sxc->c30(mu,mtop);
  double c21 = sxc->c21(mu,mtop);
  if(MSbar) c30 += 8./3.*ZETA3;
  double c1 = 2*c10;
  double c2 = 2.*c20+c11;
  double c3 = 2.*(c30+c21);
  // scale dependence
  c30 += c20*lf + c1*lf*lf/4. + lf*lf*lf/6.;
  c21 += (c2-c20)*lf + 3.*c1*lf*lf/4. + lf*lf*lf/2.;
  c20 += (c1+lf)*lf/2.;
  c2  += (c1+lf)*lf*2.;
  c1  += 2*lf;
  //
  double hard = (c3*e01*(e01*e01*pow(log(z),2)/2.-3*e01*e00*log(z)+3*e00*e00)
		 - (3*c30+c21)*2*b0*(2*e00*e01-e01*e01*log(z))
		 + 4*c30*b0*b0*e01
		 + 2*c2*(-e01*log(z)+e00)*e11
		 - 4*c20*b0*e11
		 - c1*e22*log(z) )/z;
  //
  return (hard - 12.*MomConsCoeff[2]) * pow(1-z, 2);
}
// expansion of log^2(1/z)/z in powers of (z-1) from 0 to 37
double log2z_softexp[38] = {0, 0, 1., -2., 2.916666667, -3.75, 4.511111111, -5.211111111, 5.859325397, -6.463293651, 7.029087302, -7.561626984, 8.064939875, -8.542356902, 8.996661725, -9.430203368, 9.844981992, -10.24271481, 10.62488732, -10.99279343, 11.34756740, -11.69020927, 12.02160551, -12.34254580, 12.65373676, -12.95581341, 13.24934866, -13.53486123, 13.81282242, -14.08366181, 14.34777206, -14.60551316, 14.85721599, -15.10318539, 15.34370294, -15.57902922, 15.80940597, -16.03505782 };
double log1z_softexp[38] = {0, -1., 1.5, -1.833333333, 2.083333333, -2.283333333, 2.45, -2.592857143, 2.717857143, -2.828968254, 2.928968254, -3.019877345, 3.103210678, -3.180133755, 3.251562327, -3.318228993, 3.380728993, -3.439552523, 3.495108078, -3.547739657, 3.597739657, -3.645358705, 3.690813250, -3.734291511, 3.775958178, -3.815958178, 3.854419716, -3.891456753, 3.927171039, -3.961653798, 3.994987131, -4.027245195, 4.058495195, -4.088798226, 4.118209990, -4.146781419, 4.174559197, -4.201586224};
// extra term to be added to a soft-expansion
  double FO_3_smallx(double z, double mu, int truncation, int nlogs = 1) {
  double lf = -2.*log(muF/muM);
  double c10 = sxc->c10(mu,mtop);
  double c20 = sxc->c20(mu,mtop);
  double c11 = sxc->c11(mu,mtop);
  double c30 = sxc->c30(mu,mtop);
  double c21 = sxc->c21(mu,mtop);
  if(MSbar) c30 += 8./3.*ZETA3;
  double c1 = 2*c10;
  double c2 = 2.*c20+c11;
  double c3 = 2.*(c30+c21);
  // scale dependence
  c30 += c20*lf + c1*lf*lf/4. + lf*lf*lf/6.;
  c21 += (c2-c20)*lf + 3.*c1*lf*lf/4. + lf*lf*lf/2.;
  c20 += (c1+lf)*lf/2.;
  c2  += (c1+lf)*lf*2.;
  c1  += 2*lf;
  //
  double le00 = (-11.*CA+2.*_nf*(2.*CF/CA-1.))/12.;
  double le11 = _nf*(26.*CF-23.*CA)/36.;
  double le22 = CA*( (18.*ZETA2-71.)*(CA-2*CF)*_nf + CA*CA*(54*ZETA3+99.*ZETA2-395) ) /108.;
  //
  double sub2=0, sub1=0, sub0=0;
  for(int i=0; i<=truncation; i++) {
    sub2 += log2z_softexp[i] * pow(z-1,i);
    sub1 -= log1z_softexp[i] * pow(z-1,i);
    sub0 += pow(1-z,i);
  }
  double hard = 0;
  if(nlogs>0) hard += c3*e01*e01*e01/2. * ( pow(log(z),2)/z - sub2 );
  if(nlogs>1) hard += ((3*c30+c21)*2*b0*e01*e01-c3*e01*3*e01*le00- c1*le22 -2*c2*e01*le11)* ( log(z)/z - sub1 );
  if(nlogs>2) hard += (c3*e01*3*le00*le00-(3*c30+c21)*2*b0*2*le00*e01+ 4*c30*b0*b0*e01+ 2*c2*le00*le11- 4*c20*b0*le11 ) *( 1./z - sub0 );
  //
  return hard;
}
  int TruncationOrder = 37;
  int C3mode = 2;
  void SetN3LOmode(int mode, int truncationOrder) {
    C3mode = mode;
    TruncationOrder = truncationOrder;
  }
double FO_3_nonlog(double z, double mu) {
  double nonlog = 0;
  if(finite_mt) {
    nonlog = _W1*C2reg_pointlike_gg_softexp(z, _nf, TruncationOrder) + _W2*C1reg_pointlike_gg_softexp(z, _nf, TruncationOrder) + C3reg_pointlike_gg(z, _nf, TruncationOrder);
    if(C3mode==2) nonlog += FO_3_smallx(z, mu, TruncationOrder, 1);
    if(C3mode==3) nonlog += FO_3_smallx(z, mu, TruncationOrder, 2);
    if(C3mode==4) nonlog += FO_3_smallx(z, mu, TruncationOrder, 3);
  } else {
    if(C3mode==0) nonlog = _W1*C2reg_pointlike_gg_softexp(z, _nf, TruncationOrder) + _W2*C1reg_pointlike_gg_softexp(z, _nf, TruncationOrder) + C3reg_pointlike_gg(z, _nf, TruncationOrder);
    if(C3mode==1) nonlog = _W1*(C2reg_pointlike_gg_softexp(z, _nf, TruncationOrder)+C2reg_pointlike_gg_smallx(z, _nf, TruncationOrder)) + _W2*(C1reg_pointlike_gg_softexp(z, _nf, TruncationOrder)+C1reg_pointlike_gg_smallx(z, _nf, TruncationOrder)) + C3reg_pointlike_gg(z, _nf, TruncationOrder);
    if(C3mode>=2) nonlog = _W1*C2reg_pointlike_gg(z, _nf) + _W2*C1reg_pointlike_gg(z, _nf) + C3reg_pointlike_gg(z, _nf, TruncationOrder);
    if(C3mode>=3) nonlog += C3reg_pointlike_gg_smallx(z);  // adds the 1/N^6 pole
  }
  return nonlog;
}
double FO_3_qg(double z, double mu) {
  double res = 0;
  // Vogt qg channel
  //double Pgq = CF*(1.+pow(1.-z,2))/z/2.;
  //res = 2*Pgq * pow(log(1-z)-log(z)/2.,5) * (18*CF*CF+100./3.*CF*CA+230./3.*CA*CA)/pow(4.,3);
  //
  res = _W1*C2reg_pointlike_qg(z, _nf) + _W2*C1reg_pointlike_qg(z, _nf) + C3reg_pointlike_qg(z, _nf, TruncationOrder);
  return res;
}
double FO_3_qqb(double z, double mu) {
  double res = 0;
  res = _W1*C2reg_pointlike_qqb(z, _nf) + _W2*C1reg_pointlike_qqb(z, _nf) + C3reg_pointlike_qqb(z, _nf, TruncationOrder);
  return res;
}
double FO_3_qq(double z, double mu) {
  double res = 0;
  res = _W1*C2reg_pointlike_qq(z, _nf) + C3reg_pointlike_qq(z, _nf, TruncationOrder);
  return res;
}
double FO_3_qqp(double z, double mu) {
  double res = 0;
  res = _W1*C2reg_pointlike_qqp(z, _nf) + C3reg_pointlike_qqp(z, _nf, TruncationOrder);
  return res;
}




//
// Pay attention to the appropriate definition of the soft part..
double UncertaintyCoeff[3] = {0,0,0};
void ComputeUncertaintyCoeff(double mu) {
  UncertaintyCoeff[0] = real(g0expS_order1(2.));
  UncertaintyCoeff[1] = real(g0expS_order2(2.));
  UncertaintyCoeff[2] = real(g0expS_order3(2.));
}









double FO_log(double z, double D(int,double)) {
  double res = 0;
  if(order == 1) res = FO_1_log(z,D);
  if(order == 2) res = FO_2_log(z,D);
  if(order == 3) res = FO_3_log(z,D);
  return res;
}
double FO_log_APimproved(double z, double D(int,double)) {
  double res = 0;
  if(order == 1) res = FO_1_log_APimproved(z,D);
  if(order == 2) res = FO_2_log_APimproved(z,D);
  if(order == 3) res = FO_3_log_APimproved(z,D);
  return res;
}
double FO_Nsoftcoll_impr(double z) {
  double res = 0;
  if(order == 1) res = FO_1_Nsoftcoll_impr(z);
  if(order == 2) res = FO_2_Nsoftcoll_impr(z);
  if(order == 3) res = FO_3_Nsoftcoll_impr(z);
  return res;
}
double FO_nonlog(double z, double mu) {
  double res = 0;
  if(order == 1) res = FO_1_nonlog(z,mu);
  if(order == 2) res = FO_2_nonlog(z,mu);
  if(order == 3) res = FO_3_nonlog(z,mu);
  return res;
}
double FO_hard(double z, double mu) {
  //if(!finite_mt) return 0;
  double res = 0;
  if(order == 1) res = FO_1_hard(z,mu);
  if(order == 2) res = FO_2_hard(z,mu);
  if(order == 3) res = FO_3_hard(z,mu);
  return res;
}
double FO_delta_const(double mu) {
  double res = 0;
  if(order == 1) res = FO_1_delta_const();
  if(order == 2) res = FO_2_delta_const(mu);
  if(order == 3) res = FO_3_delta_const(mu);
  return res;
}
double FO_qg(double z, double mu) {
  double res = 0;
  if(order == 1) res = FO_1_qg(z, mu);
  if(order == 2) res = FO_2_qg(z, mu);
  if(order == 3) res = FO_3_qg(z, mu);
  return res;
}
double FO_qqb(double z, double mu) {
  double res = 0;
  if(order == 1) res = FO_1_qqb(z, mu);
  if(order == 2) res = FO_2_qqb(z, mu);
  if(order == 3) res = FO_3_qqb(z, mu);
  return res;
}
double FO_qq(double z, double mu) {
  double res = 0;
  if(order == 2) res = FO_2_qq(z, mu);
  if(order == 3) res = FO_3_qq(z, mu);
  return res;
}
double FO_qqp(double z, double mu) {
  double res = 0;
  if(order == 2) res = FO_2_qqp(z, mu);
  if(order == 3) res = FO_3_qqp(z, mu);
  return res;
}












double FO_distr_gg(double z, double mu) {
  double res = 0;
  if(__exact__)
    res = FO_log(z,Dk);
  else if(__soft_part__) {
    if(__log_type__ == 2)
      res = FO_log(z,DMPk);
    else
      res = FO_log(z,Dk);
  }
  //
  //double lrf = 2.*log(_muR_muF_ratio);
  //if(order == 2) res += 3*b0*FO_1_log(z,Dk)*lrf;
  //if(order == 3) res += (3*b1+6*b0*b0*lrf)*lrf*FO_1_log(z,Dk) + 4*b0*FO_2_log(z,Dk)*lrf;
  return res;
}
double FO_reg_gg(double z, double mu) {
  double res = 0;
  if(__exact__) {
    res = FO_nonlog(z,mu);
  }
  else {
    if(__soft_part__ && __log_type__==0 && APtype!=0)
      res = FO_log_APimproved(z,Dk) - FO_log(z,Dk);
    if(__soft_part__ && __log_type__==1 && APtype!=0)
      res = FO_log_APimproved(z,Dtk) - FO_log(z,Dk);
    if(__soft_part__ && __log_type__==2 && __Nsoftcoll__)
      res = FO_Nsoftcoll_impr(z);
    if(__hard_part__)
      res += FO_hard(z,mu);
  }
  /*
  if(__exact__ || __soft_part__) {
    double lrf = 2.*log(_muR_muF_ratio);
    if(order == 2) res += 3*b0*FO_1_nonlog(z,mu)*lrf;
    if(order == 3) res += (3*b1+6*b0*b0*lrf)*lrf*FO_1_nonlog(z,mu) + 4*b0*FO_2_nonlog(z,mu)*lrf;
  }
  */
  return res;
}
double FO_delta_gg(double mu) {
  double res = 0;
  if(__exact__ || __soft_part__) {
    res = FO_delta_const(mu);
    if(__log_type__ == 2) res += FO_log(0, mell_const);
    /*
    double lrf = 2.*log(_muR_muF_ratio);
    if(order == 1) res += 2*b0*lrf;
    if(order == 2) res += (2*b1+3*b0*FO_1_delta_const())*lrf + 3*b0*b0*lrf*lrf;
    if(order == 3) res += (2*b2+3*b1*FO_1_delta_const()+4*b0*FO_2_delta_const(mu))*lrf
		     + (7*b0*b1+6*b0*b0*FO_1_delta_const())*lrf*lrf + 4*b0*b0*b0*lrf*lrf*lrf;
    */
  }
  return res;
}






// gg channel
// integrand of regular part
double gg_reg_int(double t1, double t2, double tau, double mu) {
  double z = 1.-(1.-tau)*t1;
  double res = lumi(lumi_gg, t2, tau/z, muFpdf) * FO_reg_gg(z, mu) /z;
  return res * (1-tau);
}
// integrand of distributional part
double gg_distr_int(double t1, double t2, double tau, double mu) {
  double lumtau = lumi(lumi_gg, t2, tau, muFpdf);
  double res = 0;//lumtau * FO_delta_gg(mu);  // delta term
  if(order>0) {
    double z = 1.-(1.-tau)*t1;
    res += (1-tau) * ( lumi(lumi_gg, t2, tau/z, muFpdf)/z - lumtau ) * FO_distr_gg(z, mu);
    res += - lumtau * FO_distr_gg(t1*tau, mu) * tau; // add residual integral form 0 to tau
  }
  return res;
}
//
// non-gg channels contributions
double qg_int(double t1, double t2, double tau, double mu) {
  double z = 1.-(1.-tau)*t1;
  double res = lumi(lumi_qg, t2, tau/z, muFpdf) * FO_qg(z, mu) /z;
  return res * (1-tau);
}
//
  /*
double qqb_int(double t1, double t2, double tau, double mu) {
  double z = 1.-(1.-tau)*t1;
  double res = lumi(lumi_qqb, t2, tau/z, muFpdf) * FO_qqb(z, mu) /z;
  return res * (1-tau);
}
//
double qq_int(double t1, double t2, double tau, double mu) {
  double z = 1.-(1.-tau)*t1;
  double res = lumi(lumi_qq, t2, tau/z, muFpdf) * FO_qq(z, mu) /z;
  return res * (1-tau);
}
//
double qqp_int(double t1, double t2, double tau, double mu) {
  double z = 1.-(1.-tau)*t1;
  double res = lumi(lumi_qqp, t2, tau/z, muFpdf) * FO_qqp(z, mu) /z;
  return res * (1-tau);
}
  */
double quark_channels_int(double t1, double t2, double tau, double mu) {
  double z = 1.-(1.-tau)*t1;
  double res = 0;
  res += lumi(lumi_qqb, t2, tau/z, muFpdf) * FO_qqb(z, mu) /z;
  if(order>1) {
    res += lumi(lumi_qq , t2, tau/z, muFpdf) * FO_qq (z, mu) /z;
    res += lumi(lumi_qqp, t2, tau/z, muFpdf) * FO_qqp(z, mu) /z;
  }
  return res * (1-tau);
  // version with non-singlet basis
  double qqb = FO_qqb(z, mu);
  double qq  = FO_qq (z, mu);
  double qqp = FO_qqp(z, mu);
  double FO_SS = (qq + qqb + 2*(_nf-1.)*qqp)/(2.*_nf);
  double FO_TT = (qq + qqb - 2*qqp)/2.;
  double FO_VV = (qq - qqb)/2.;
  res += lumi(lumi_SS, t2, tau/z, muFpdf) * FO_SS /z;
  res += lumi(lumi_VV, t2, tau/z, muFpdf) * FO_VV /z;
  res += lumi(lumi_TT, t2, tau/z, muFpdf) * FO_TT /z;
  return res * (1-tau);
}




struct int1dp { double tau, mu; };
bool __all_channels__ = false;
int integrandCUBA(const int *ndim, const double x[], const int *ncomp, double res[], void *pars) {
  int1dp p = *(int1dp *) pars;
  double tau = p.tau;
  double mu  = p.mu;
  if(finite_mt && order==1) _integration_v_ = x[2];
  res[0] = 0;
  if(_gg_channel_) res[0] += gg_reg_int(x[0],x[1],tau,mu);
  if(__exact__ || __soft_part__)
    if(_gg_channel_) res[0] += gg_distr_int(x[0],x[1],tau,mu);
  if(__all_channels__ && __exact__) {
    if(_qg_channel_) res[0] += qg_int(x[0],x[1],tau,mu);
    if(_qq_channel_) res[0] += quark_channels_int(x[0],x[1],tau,mu);
    //res[0] += qqb_int(x[0],x[1],tau,mu);
    //if(order > 1) res[0] += qq_int(x[0],x[1],tau,mu) + qqp_int(x[0],x[1],tau,mu);
  }
  return 0;
}
double integral2d(double tau, double mu, double *err = NULL) {
  double res = 0;
  int1dp p = { tau, mu };
  //
  double cres[1], error[1], prob[1];
  double epsrel=INTEGRAL_PRECISION, epsabs=1.e-10;
  int last = 4;
  int verbose = 0;
  int nregions, neval, fail;
  //system("export CUBACORES=8");
  int dim = 2;
  if(finite_mt && order==1) dim = 3;
  Cuhre(dim, 1, integrandCUBA, &p, 1,
	epsrel, epsabs, verbose | last,
	0, 500000, 9,
	NULL, NULL,
	&nregions, &neval, &fail, cres, error, prob);
  // Suave(2, 1, integrandCUBA, &p,
  // 	epsrel, epsabs, verbose | last, 0,
  // 	0, 500000, 1000, 25.,
  // 	&nregions, &neval, &fail, cres, error, prob);
  // Vegas(2, 1, integrandCUBA, &p,
  // 	epsrel, epsabs, verbose | last, 0,
  // 	0, 500000, 1000, 500, 1000,
  // 	0, NULL,
  // 	&neval, &fail, cres, error, prob);
  //  
  if(fail == -1) {
    cout << "\033[0;31m  ERROR: dimension out of range.\033[0m" << endl;
  } else if(fail>0) {
    cout << "\033[0;31m  ERROR: precision not reached.\033[0m" << endl;
  }
  double lum_tau = Lum(tau, muFpdf);
  res += cres[0]/lum_tau;
  if(err != NULL) *err = error[0]/lum_tau;
  return res;
}


double xs(double M, double sqrts, double *err = NULL) {
  double tau = M*M/sqrts/sqrts;
  double res = integral2d(tau, M, err);
  if(_gg_channel_) res += FO_delta_gg(M);
  return res;
}




dcomplex brn_q_ep0(dcomplex x) {
  dcomplex logx = log(x);
  dcomplex HPL2 = 0.5*logx*logx;
  return -16.*(1.+x)*(1.+x)/x*HPL2 + 32.*(x-1.)*(x-1.)/x;
}
dcomplex sborn_mass(dcomplex mH2, vector<double> m) {
  dcomplex res = 0;
  for(int i=0; i<m.size(); i++) {
    dcomplex Wq = 4.*m[i]*m[i]/mH2;
    dcomplex Xq = -Wq/pow(1.+sqrt(1.-Wq),2);
    res += brn_q_ep0(Xq)*Wq*Wq /64.*(-3./4.);
  }
  return res;
}
double xs_LO_prefactor_EFT() {
  double pref = 35.0309/M_PI/M_PI; // = Gf/pi/sqrt(2)/288 with the Gf in pb
  return pref;
}
double EFTrescaling(double mH, double mt, double mb, double mc) {
  vector<double> m;
  if(mt>0) m.push_back(mt);
  if(mb>0) m.push_back(mb);
  if(mc>0) m.push_back(mc);
  dcomplex mH2 = mH*mH + I*1e-10; // analytic continuation
  return norm(sborn_mass(mH2,m));
}
double xs_LO_prefactor(double mH) {
  return xs_LO_prefactor_EFT() * EFTrescaling(mH,mtop,mbot,mcha);
}
  double EFTrescaling_deriv(double mH, double mt, double mb, double mc, int id, double &second_deriv, double &third_deriv) {
    double m0[3] = {mt, mb, mc};
    double m [3] = {mt, mb, mc};
    double eps = 0.2; // cross-checked with Mathematica for the top - agreement within 6 digits even for the third derivative
    const int np = 9;
    double res[np];
    for(int i=0; i<np; i++) {
      m [id] = m0[id] + eps*(i-(np-1)/2);
      res[i] = EFTrescaling(mH,m[0],m[1],m[2]);
    }
    //double fd_coeff[np] = { -1./60., 3./20., -3./4., 0, 3./4., -3./20., 1./60. };
    //double sd_coeff[np] = { 1./90., -3./20., 3./2., -49./18., 3./2., -3./20., 1./90. };
    //double td_coeff[np] = { 1./8., -1., 13./8., 0, -13./8., 1., -1./8. };
    double fd_coeff[np] = { 1./280., -4./105., 1./5., -4./5., 0, 4./5., -1./5., 4./105., -1./280. };
    double sd_coeff[np] = { -1./560., 8./315., -1./5., 8./5., -205./72., 8./5., -1./5., 8./315., -1./560. };
    double td_coeff[np] = { -7./240., 3./10., -169./120., 61./30., 0, -61./30., 169./120., -3./10., 7./240. };
    double first_deriv = 0;
    second_deriv = 0;
    third_deriv = 0;
    for(int i=0; i<np; i++) {
      first_deriv  += fd_coeff[i] * res[i];
      second_deriv += sd_coeff[i] * res[i];
      third_deriv  += td_coeff[i] * res[i];
    }
    first_deriv  /= eps;
    second_deriv /= eps*eps;
    third_deriv  /= eps*eps*eps;
    return first_deriv;
  }

  double EFT_rescaling_MSbar_NLO(double mH) {
    double fd, sd, td;
    fd = EFTrescaling_deriv(mH, mtop, mbot, mcha, 0, sd, td);
    return fd * mtop * alphas(mut)/M_PI * (-4./3.-2*log(mut/mtmt));
  }





bool firstcall = true;
void compute(double mH, double sqrts, double LO, double NLO, double NNLO,
	     double *rexact, double *rsoft2, double *rsoft1, double *rNsoft,
	     double *rhighenergy, double *momconsunc, double *rhighenergyLL,
	     double *rsoft0, double *rsoftAP, double *rNsoftcoll, double *rNNLL, double *rNsoftdFMMV) {
  qmasses.clear();
  if(mtop>0) qmasses.push_back(mtop);
  if(mbot>0) qmasses.push_back(mbot);
  if(mcha>0) qmasses.push_back(mcha);
  mH2 = mH*mH + I*1e-10; // analytic continuation
  double as = alphas(mH*_muF_mu_ratio*_muR_muF_ratio); // as(muR)
  //cout << "alpha_s = "<< as << endl;
  double api = as/M_PI;
  double fact = pow(api,order) * LO;
  double tau = mH*mH/sqrts/sqrts;
  SetScales(mH);
  init_consts(mH);
  ApproxLuminosity(ChebCoeff_gg, tau, Nchebyshev, MPbeta, MPgamma, 0);
  ApproxLuminosity(ChebCoeff_qg, tau, Nchebyshev, MPbeta, MPgamma, 1);
  ApproxLuminosity(ChebCoeff_SS, tau, Nchebyshev, MPbeta, MPgamma, 2);
  ApproxLuminosity(ChebCoeff_TT, tau, Nchebyshev, MPbeta, MPgamma, 3);
  ApproxLuminosity(ChebCoeff_VV, tau, Nchebyshev, MPbeta, MPgamma, 4);
  /*
  LumChannel = 4;
  for(int i=0; i<20; i++) {
    double z = 0.999*exp(i/(20-1.)*log(tau));
    double l1 = pow(z,MPbeta)/pow(1-z,MPgamma)*ChebLum(z);
    double l2 = pow(z,MPbeta)/pow(1-z,MPgamma)*ChebRecLum(ChebCoeff_VV, Nchebyshev, z, MPbeta, MPgamma);
    cout << setw(12) << z
	 << setw(15) << l1
	 << setw(15) << l2
	 << setw(15) << (l2-l1)/l1
	 << endl;
  }
  abort();
  //*/
  //
  if(!quiet && firstcall) cout << "sigma = sigma0 * [ 1 + as K1 + as^2 K2 + as^3 K3 + ...]" << endl;
  firstcall = false;
  //
  string channel = (__all_channels__ ? "all channels:" : "gg channel:");
  if(!quiet) cout << channel << endl;
  string sassig[4] = { "  sig0 = ",
		       "  as   K1 = ",
		       "  as^2 K2 = ",
		       "  as^3 K3 = " };
  double err;
  double lrf = 2.*log(_muR_muF_ratio);
  double lr  = 2.*log(_muR_muF_ratio*_muF_mu_ratio);
  double mur_dep_term = 0; // using terms of previous orders *at* mur
  if(!_gg_channel_) LO = 0;
  NNLO -= NLO;
  NLO -= LO;
  if(exactmuRdep) {
    if(order == 1) mur_dep_term = api*LO*2*b0*lrf;
    if(order == 2) mur_dep_term = api*(api*2*b1*LO + 3*b0*NLO) *lrf - api*api*3*b0*b0*LO*lrf*lrf;
    //if(order == 2) mur_dep_term = api*(api*2*b1*LO + 3*b0*NLO) *lr - api*api*3*b0*b0*LO*lr*lr;
    if(order == 3) mur_dep_term = api*(api*api*2*b2*LO + api*3*b1*NLO + 4*b0*NNLO) *lrf
    		     - api*api*(api*7*b0*b1*LO + 6*b0*b0*NLO) *lrf*lrf + api*api*api*4*b0*b0*b0*LO*lrf*lrf*lrf;
    if(newevolmuFmuR)
      if(order == 3) mur_dep_term = api*(api*api*2*b2*LO + api*3*b1*NLO + 4*b0*NNLO) *lr
		       - api*api*(api*7*b0*b1*LO + 6*b0*b0*NLO) *lr*lr + api*api*api*4*b0*b0*b0*LO*lr*lr*lr;
  }
  //cout << order << "  " << mur_dep_term << endl;
  //
  computesoft = false;
  __exact__ = true; // compute the full gg contribution to the hadronic xs
  __log_type__ = 0;
  APtype = 0;
  if(rexact != NULL) {
    int fromorder = 3;
    double mufrat = _muF_mu_ratio;
    if(order>=fromorder) _muF_mu_ratio = 1;
    *rexact = xs(mH, sqrts, &err) * fact + mur_dep_term;
    _muF_mu_ratio = mufrat;
    if(order>=fromorder) {
      //cout << *rexact/fact << endl;
      *rexact += MellinInversion(tau, &err) * fact;
    }
    err *= fact;
    if(!quiet) cout << sassig[order] << *rexact/LO << " +/- " << err/LO
		    << "  (exact)" << endl;
  }
  //
  computesoft = true;
  __exact__ = false; // compute only approximations
  __soft_part__ = true; // turn on soft
  __hard_part__ = false; // leave off hard
  __log_type__ = 0;
  APtype = 0;
  //APtype = 1;  // reproduces NkLO(sv) of 1510.02235
  if(rsoft0 != NULL) {
    bool APimpExp = APimprExponent;
    APimprExponent = false;
    *rsoft0 = xs(mH, sqrts, &err) * fact + mur_dep_term;
    //if(APimprExponent)
    //  *rsoft0 = MellinInversion(tau, &err) * fact + mur_dep_term;
    //else
    //  *rsoft0 = xs(mH, sqrts, &err) * fact + mur_dep_term;
    err *= fact;
    if(!quiet) cout << sassig[order] << *rsoft0/LO << " +/- " << err/LO
		    << "  (z-soft only)" << endl;
    APimprExponent = APimpExp;
  }
  //
  __log_type__ = 1;
  APtype = 2;
  if(rsoft2 != NULL) {
    if(APimprExponent)
      *rsoft2 = MellinInversion(tau, &err) * fact + mur_dep_term;
    else
      *rsoft2 = xs(mH, sqrts, &err) * fact + mur_dep_term;
    err *= fact;
    if(!quiet) cout << sassig[order] << *rsoft2/LO << " +/- " << err/LO
		    << "  (A-soft_2 only)" << endl;
  }
  //
  __log_type__ = 1;
  APtype = 1;
  if(rsoft1 != NULL) {
    if(APimprExponent)
      *rsoft1 = MellinInversion(tau, &err) * fact + mur_dep_term;
    else
      *rsoft1 = xs(mH, sqrts, &err) * fact + mur_dep_term;
    err *= fact;
    if(!quiet) cout << sassig[order] << *rsoft1/LO << " +/- " << err/LO
		    << "  (A-soft_1 only)" << endl;
  }
  //
  __log_type__ = 1;
  APtype = -1;
  if(rsoftAP != NULL) {
    if(APimprExponent)
      *rsoftAP = MellinInversion(tau, &err) * fact + mur_dep_term;
    else
      *rsoftAP = xs(mH, sqrts, &err) * fact + mur_dep_term;
    err *= fact;
    if(!quiet) cout << sassig[order] << *rsoftAP/LO << " +/- " << err/LO
		    << "  (soft_inf (full AP) only)" << endl;
  }
  //
  APtype = 0;
  if(rNsoft != NULL) {
    __log_type__ = 3;
    *rNsoft = MellinInversion(tau, &err) * fact + mur_dep_term;
    //__log_type__ = 2;
    //*rNsoft = xs(mH, sqrts, &err) * fact + mur_dep_term;
    err *= fact;
    if(!quiet) cout << sassig[order] << *rNsoft/LO << " +/- " << err/LO
		    << "  (N-soft only)" << endl;
  }
  //
  if(rNsoftdFMMV != NULL) {
    dFMMV = true;
    __log_type__ = 3;
    *rNsoftdFMMV = MellinInversion(tau, &err) * fact + mur_dep_term;
    err *= fact;
    if(!quiet) cout << sassig[order] << *rNsoftdFMMV/LO << " +/- " << err/LO
		    << "  (N-soft only)" << endl;
    dFMMV = false;
  }
  //
  __log_type__ = 2;
  if(rNsoftcoll != NULL) {
    __Nsoftcoll__ = true;
    *rNsoftcoll = xs(mH, sqrts, &err) * fact + mur_dep_term;
    err *= fact;
    if(!quiet) cout << sassig[order] << *rNsoftcoll/LO << " +/- " << err/LO
		    << "  (N-soft-collinear only)" << endl;
    __Nsoftcoll__ = false;
  }
  //
  if(rNNLL != NULL) {
    __NNLL__ = true;
    *rNNLL = xs(mH, sqrts, &err) * fact + mur_dep_term;
    err *= fact;
    if(!quiet) cout << sassig[order] << *rNNLL/LO << " +/- " << err/LO
		    << "  (N-soft-collinear only)" << endl;
    __NNLL__ = false;
  }
  //
  __soft_part__ = false; // turn soft off
  __hard_part__ = true;  // turn hard on
  initGammaCoeffs(true, _nf);
  //double mufrat = _muF_mu_ratio;
  //if(order>=3) _muF_mu_ratio = 1;
  if(rhighenergy != NULL) {
    // momentum conservation
    ComputeMomConsCoeff(mH);
    *rhighenergy = xs(mH, sqrts, &err) * fact;
    err *= fact;
    if(!quiet) cout << sassig[order] << *rhighenergy/LO << " +/- " << err/LO
		    << "  (high-energy --- must be added to a soft)" << endl;
  }
  //
  if(momconsunc != NULL) {
    if(rhighenergy != NULL) {
      *momconsunc = *rhighenergy;
    } else {
      ComputeMomConsCoeff(mH);
      *momconsunc = xs(mH, sqrts, &err) * fact;
    }
    double MCcoeff = MomConsCoeff[order-1];
    ResetMomConsCoeff();
    *momconsunc -= xs(mH, sqrts, &err) * fact; // mom-cons function
    //
    APtype = 2;  __log_type__ = 1;
    ComputeUncertaintyCoeff(mH);
    //cout << UncertaintyCoeff[order-1]/MCcoeff << endl;
    //
    *momconsunc *= -UncertaintyCoeff[order-1]/MCcoeff;
    //cout << *momconsunc << endl;
    err *= fact*UncertaintyCoeff[order-1]/MCcoeff;
    if(!quiet) cout << sassig[order] << *momconsunc*0.05/LO << " +/- " << err*0.05/LO
		    << "  (momcons uncertainty)" << endl;
  }
  initGammaCoeffs(false, _nf);
  if(rhighenergyLL != NULL) {
    // momentum conservation
    ComputeMomConsCoeff(mH);
    *rhighenergyLL = xs(mH, sqrts, &err) * fact;
    err *= fact;
    if(!quiet) cout << sassig[order] << *rhighenergyLL/LO << " +/- " << err/LO
		    << "  (high-energy at LL)" << endl;
  }
  //_muF_mu_ratio = mufrat;
}



  double GetZmass() {
    return mZ;
  }


};

double ChebLum(double z, int LumChannel) {
  return ggHiggs::Lumij(z, ggHiggs::muFpdf, LumChannel);
}


